'''
Predefined queries for Cirrus API

Designed to work with the following schema (2020-01-17):
https://github.com/cirruslabs/cirrus-ci-web/blob/1806cccc/schema.graphql
'''


from time import monotonic as time, sleep
import logging

from . import CirrusAPI


log = logging.getLogger(__name__)


class CirrusQueryError(ValueError):
    '''Raised when query executes successfully but returns invalid data'''


class CirrusBuildError(RuntimeError):
    '''Raised on build failures'''


class CirrusTimeoutError(RuntimeError):
    '''Raised when build takes too long'''


def get_repo(api: CirrusAPI, owner: str, repo: str) -> str:
    '''Get internal ID for GitHub repo'''
    query = '''
        query GetRepos($owner: String!) {
            githubRepositories(owner: $owner) {
                id
                name
            }
        }
    '''
    params = dict(owner=owner)
    all_repos = api(query, params)
    for item in all_repos['githubRepositories']:
        if item['name'] == repo:
            return item['id']
    raise CirrusQueryError('repo not found: {}/{}'.format(owner, repo))


def create_build(api: CirrusAPI,
                 repo_id: str,
                 repo_branch: str = 'master',
                 config: str = '') -> str:
    '''
    Trigger new build on Cirrus CI

    Return build ID
    '''
    query = '''
        mutation ScheduleCustomBuild($config: String!,
                                     $repo: ID!,
                                     $branch: String!,
                                     $mutation_id: String!) {
            createBuild(
                input: {
                    repositoryId: $repo,
                    branch: $branch,
                    clientMutationId: $mutation_id,
                    configOverride: $config
                }
            ) {
                build {
                    id
                    status
                }
            }
        }
    '''
    mutation_id = 'cirrus-run job {}'.format(int(time()))
    answer = api(
        query=query,
        params=dict(
            repo=repo_id,
            branch=repo_branch,
            mutation_id=mutation_id,
            config=config),
    )
    return answer['createBuild']['build']['id']


def wait_build(api, build_id: str, delay=3, abort=60*60):
    '''Wait until build finishes'''
    ERROR_CONFIRM_TIMES = 3

    query = '''
        query GetBuild($build: ID!) {
            build(id: $build) {
                status
            }
        }
    '''
    params = dict(build=build_id)

    errors_confirmed = 0
    time_start = time()
    while time() < time_start + abort:
        response = api(query, params)
        status = response['build']['status']
        log.info('build {}: {}'.format(build_id, status))
        if status in {'COMPLETED'}:
            return True
        if status in {'CREATED', 'TRIGGERED', 'EXECUTING'}:
            errors_confirmed = 0
            sleep(delay)
            continue
        if status in {'NEEDS_APPROVAL', 'FAILED', 'ABORTED', 'ERRORED'}:
            errors_confirmed += 1
            if errors_confirmed < ERROR_CONFIRM_TIMES:
                sleep(2 * delay / (ERROR_CONFIRM_TIMES - 1))
                continue
            else:
                raise CirrusBuildError('build {} was terminated: {}'.format(build_id, status))
        raise ValueError('build {} returned unknown status: {}'.format(build_id, status))
    raise CirrusTimeoutError('build {} timed out'.format(build_id))


def build_log(api, build_id):
    '''Yield build log in chunks of text'''
    query = '''
        query GetBuildLog($build: ID!) {
            build(id: $build) {
                tasks {
                    id
                    name
                    commands {
                        name
                    }
                }
            }
        }
    '''
    params = dict(build=build_id)
    url_template = 'https://api.cirrus-ci.com/v1/task/{task[id]}/logs/{command[name]}.log'
    response = api(query, params)
    for task in response['build']['tasks']:
        yield '\n## Task: {task[name]}'.format(**locals())
        for command in task['commands']:
            yield '\n## Task instruction: {command[name]}'.format(**locals())
            url = url_template.format(**locals())
            log = api.get(url)
            if log.status_code == 200:
                yield log.text
            else:
                yield 'Unable to fetch url: {}'.format(url)



'''
Command line interface for cirrus-run
'''


import argparse
import logging
import os
import sys
import traceback
from pprint import pformat

from jinja2 import Template

from . import CirrusAPI
from .throbber import ProgressBar
from .queries import build_log, get_repo, create_build, wait_build, CirrusBuildError

log = logging.getLogger(__name__)


ENVIRONMENT = {
    'github': 'CIRRUS_GITHUB_REPO',
    'branch': 'CIRRUS_GITHUB_BRANCH',
    'token': 'CIRRUS_API_TOKEN',
    'config': 'CIRRUS_CONFIG',
    'timeout': 'CIRRUS_TIMEOUT',
    'show_log': 'CIRRUS_SHOW_BUILD_LOG',
}


def main(*a, **ka):
    args = parse_args(*a, **ka)
    configure_logging(args.verbose)
    log.debug('Parsed command line arguments:\n{}'.format(pformat(vars(args), indent=2)))

    config = read_config(args.config)
    api = CirrusAPI(args.token)
    repo_id = get_repo(api, args.owner, args.repo)
    build_id = create_build(api, repo_id, args.branch, config)
    build_url = 'https://cirrus-ci.com/build/{}'.format(build_id)

    print('Build created: {}'.format(build_url))
    with ProgressBar('' if args.verbose else '.'):
        try:
            wait_build(api, build_id, abort=args.timeout*60)
            rc, status, message = 0, 'successful', ''
        except CirrusBuildError:
            rc, status, message = 1, 'failed', ''
        except Exception as exc:
            rc, status, message = 2, 'error', '{exception}: {text}'.format(
                                        exception=exc.__class__.__name__,
                                        text=str(exc))

    if args.show_build_log == 'always' \
    or (args.show_build_log == 'failure' and rc != 0):
        print('Build {}, see log below:'.format(status, build_url))
        try:
            for chunk in build_log(api, build_id):
                print(chunk)
        except Exception as exc:
            error = traceback.format_exc()
            log.error(error)

    print('Build {}: {}'.format(status, build_url))
    if message:
        print('  {}'.format(message))
    sys.exit(rc)


def read_config(path):
    '''Load YAML config from file or Jinja2 template'''
    with open(path) as config_file:
        raw = config_file.read()
    ext = os.path.splitext(path)[1].lower().lstrip('.')
    if ext in {'j2', 'jinja', 'jinja2'}:
        template = Template(raw)
        return template.render(os.environ)
    else:
        return raw


def fallback_config_path():
    '''Calculate default config path if none provided by user'''
    paths = [
        '.cirrus.yml',
        '.cirrus.yml.j2',
    ]
    for path in paths:
        if os.path.isfile(path):
            return path
    else:
        return paths[0]


def parse_args(*a, **ka):
    parser = argparse.ArgumentParser(
        description=(
            'Execute CI jobs in CirrusCI'
        ),
        epilog='Licensed under the Apache License, version 2.0',
    )
    parser.add_argument(
        'config',
        metavar='CONFIG',
        default=os.getenv(ENVIRONMENT['config'], fallback_config_path()),
        nargs='?',
        help=(
            'Path to YAML configuration file or Jinja2 template for such file. '
            'Filenames ending with .j2 or .jinja2 are assumed to provide the templates. '
            'All environment variables are available inside these templates. '
            'Default value: ${} or .cirrus.yml or .cirrus.yml.j2'
        ).format(ENVIRONMENT['config']),
    )
    parser.add_argument(
        '--token',
        default=os.getenv(ENVIRONMENT['token']),
        metavar='TOKEN',
        help=(
            'Access token for CirrusCI API. '
            'Recommended and more secure way of providing the token is via '
            'environment variable. '
            'Default value: ${}'
        ).format(ENVIRONMENT['token']),
    )
    parser.add_argument(
        '--github',
        default=os.getenv(ENVIRONMENT['github'], ''),
        metavar='REPO',
        help=(
            'GitHub repo id that will own the build ("owner/reponame"). '
            'This repo may have no relation to the CI job being executed. '
            'It may even be empty. '
            'Default value: ${}'
        ).format(ENVIRONMENT['github']),
    )
    parser.add_argument(
        '--branch',
        default=os.getenv(ENVIRONMENT['branch'], 'master'),
        metavar='BRANCH',
        help=(
            'GitHub repo branch that will own the build. '
            'This branch may have no relation to the CI job being executed. '
            'Default value: ${} or master'
        ).format(ENVIRONMENT['branch']),
    )
    parser.add_argument(
        '--owner',
        default='',
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        '--repo',
        default='',
        help=argparse.SUPPRESS,
    )
    parser.add_argument(
        '-v',
        '--verbose',
        action='count',
        default=0,
        help=('Increase output verbosity. Repeating this argument multiple times '
              'increases verbosity level even further.'),
    )
    parser.add_argument(
        '--timeout',
        default=os.getenv(ENVIRONMENT['timeout'], 120),
        metavar='MINUTES',
        help=(
            'Timeout (in minutes) before assuming that the build has hanged and '
            'that API responses are unreliable. Default value: ${} or 120'
        ).format(ENVIRONMENT['timeout']),
    )
    parser.add_argument(
        '--show-build-log',
        default=os.getenv(ENVIRONMENT['show_log'], 'failure'),
        choices={'always', 'never', 'failure'},
        type=str.lower,
        help=(
            'Specify whether to print the build log to stdout after completing CI run. '
            'Default value: ${} or "failure"'
        ).format(ENVIRONMENT['show_log']),
    )
    args = parser.parse_args(*a, **ka)

    if not args.token:
        parser.error('API token is not defined')

    if not args.github:
        parser.error('GitHub repo is not defined')

    repo_parts = args.github.split('/')
    if len(repo_parts) != 2 or not all(repo_parts):
        parser.error('invalid repo identifier: {}'.format(args.github))
    args.owner, args.repo = repo_parts

    if not os.path.isfile(args.config):
        parser.error('config file not found: {}'.format(args.config))

    return args


def configure_logging(verbosity):
    verbosity_levels = {
        0: logging.WARNING,
        1: logging.INFO,
        2: logging.DEBUG,
        3: logging.NOTSET + 1,
    }
    if verbosity > max(verbosity_levels):
        verbosity = max(verbosity_levels)
    level = verbosity_levels.get(verbosity)
    log = logging.getLogger(__name__.split('.')[0])
    log.level = min(log.level, level)


if __name__ == '__main__':
    print(parse_args())



'''
CirrusCI API interaction
'''


import logging
from time import sleep
from pprint import pformat

import requests


log = logging.getLogger(__name__)


class CirrusAPIError(Exception):
    def __init__(self, errors):
        message = 'API returned {num} error(s):\n{errors}'.format(
            num=len(errors),
            errors=pformat(errors, indent=2)
        )
        super().__init__(message)


class CirrusHTTPError(Exception):
    def __init__(self, response):
        code = response.status_code
        text = response.text
        url = response.url
        message = 'HTTP {code}: {url}\n{text}'.format(**locals())
        super().__init__(message)
        self.code = code
        self.response = response

    def __repr__(self):
        return '<{cls} [{code}]>'.format(cls=self.__class__.__name__, code=self.code)


class CirrusAPI:
    '''Interact with Cirrus via GraphQL API'''

    DEFAULT_URL = 'https://api.cirrus-ci.com/graphql'
    USER_AGENT = 'Cirrus remote CLI <https://github.com/sio/cirrus-run>'
    RETRY_ATTEMPTS = 3
    RETRY_DELAY = 2  # seconds
    RETRY_LONG_DELAY = 30

    def __init__(self, token, url=None):
        if url is None:
            url = self.DEFAULT_URL
        self._url = url

        session = requests.Session()
        session.headers.update({
            'Accept': 'application/json',
            'Authorization': 'Bearer {}'.format(token),
            'User-Agent': self.USER_AGENT,
        })
        session.timeout = 5
        self._requests = session

    def __call__(self, query, params=None, retries=None, delay=None):
        if retries is None:
            retries = self.RETRY_ATTEMPTS
        if delay is None:
            delay = self.RETRY_DELAY

        payload = dict(query=query.strip(), variables=params or {})
        log.debug('Calling API with: {}'.format(payload))

        error_count = 0
        long_wait_happened = False
        while True:
            try:
                answer = self._post(json=payload)
                return self._parse_api_response(answer)
            except Exception as exc:
                error_count += 1
                if error_count > retries:
                    raise exc
                elif isinstance(exc, CirrusHTTPError) \
                and 'try again in 30 seconds' in exc.response.text \
                and not long_wait_happened:
                    long_wait_happened = True
                    log.debug('API server asked for longer retry delay: {}, retrying'.format(exc))
                    sleep(self.RETRY_LONG_DELAY)
                else:
                    log.debug('Error when calling API: {}, retrying'.format(exc))
                    sleep(delay)

    def _parse_api_response(self, data):
        if 'errors' in data and data['errors']:
            raise CirrusAPIError(data['errors'])
        return data['data']

    def _post(self, **ka):
        response = self._requests.post(self._url, **ka)
        if response.status_code != 200:
            raise CirrusHTTPError(response)
        return response.json()

    def get(self, *a, **ka):
        '''Perform GET request using API session'''
        return self._requests.get(*a, **ka)



'''
Logging setup
'''

import logging
import os


class ShyLogHandler(logging.StreamHandler):
    '''Emit messages only if root logger has no other handlers'''
    root = logging.getLogger()
    def emit(self, record):
        if not self.root.hasHandlers():
            super().emit(record)


def setup():
    package = __name__.split('.')[0]
    log = logging.getLogger(package)
    if getattr(log, 'initialized', False):
        return log

    log_file = os.environ.get('CIRRUS_LOG_FILE')
    log_stdout = True

    log.level = logging.WARNING

    if log_file:
        handler = logging.FileHandler(log_file, encoding='utf-8')
        handler.level = logging.DEBUG
        handler.formatter = logging.Formatter('%(asctime)s - %(message)s')
        log.addHandler(handler)

    if log_stdout:
        handler = ShyLogHandler()
        handler.level = logging.DEBUG
        log.addHandler(handler)

    log.initialized = True
    return log



'''
Show progress in terminal
'''


from threading import Thread
from time import sleep


class ProgressBar:
    '''Simple progress indicator for long running CLI task'''

    def __init__(self, char='.', step=1, break_line=60):
        self.char = char
        self.step = step
        self.break_line = break_line
        self.exit = False
        self.thread = Thread(target=self.show)

    def __enter__(self):
        if self.char:
            self.thread.start()

    def __exit__(self, *a, **ka):
        self.exit = True
        if self.char:
            self.thread.join()

    def show(self):
        cutoff = int(self.break_line / self.step)
        ticks = 0
        exit = False
        while not exit:
            exit = self.exit
            end = ''
            if exit:
                end = '\n'
            ticks += 1
            if ticks == cutoff:
                ticks = 0
                end = '\n'
            self.tick(end)
            if not exit:
                sleep(self.step)

    def tick(self, end=''):
        print(self.char, end=end, flush=True)




VERSION='0.4.0'


from setuptools import setup, find_packages


with open("README.md") as readme:
    long_description = readme.read()


setup(
    name='cirrus-run',
    version=VERSION,
    description='Command line tool to execute jobs in Cirrus CI',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sio/cirrus-run',
    download_url='https://github.com/sio/cirrus-run/archive/v{}.tar.gz'.format(VERSION),
    author='Vitaly Potyarkin',
    author_email='sio.wtf@gmail.com',
    license='Apache-2.0',
    platforms='any',
    entry_points={
        'console_scripts': [
            'cirrus-run=cirrus_run.cli:main',
        ],
    },
    packages=find_packages(exclude=('tests',)),
    include_package_data=True,
    install_requires=[
        'Jinja2',
        'requests',
    ],
    extras_require={ },
    python_requires='>=3.4',
    zip_safe=True,
    keywords=[
        'api',
        'ci',
        'cirrus-ci',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Quality Assurance',
        'Topic :: Software Development',
    ],
)
